#Module for buildings and building manager

import math #Used for rounding up using ceil()
from collections import defaultdict #dictionary that returns default value when the value isn't found

class BasicBuilding():
    def __init__(self, parent, name, x, y, size, hp, weapon, tool, cargo_capacity, time_to_finish = 0, tool_to_build = ""):
        self.parent = parent
        self.name = name
        self.id = -1 #id is -1 until it is changed to real id
        
        self.x = x
        self.y = y
        self.size = size

        self.hp = hp
        self.weapon = weapon
        self.tool = tool

        self.disabled = False

        self.reload = 0
        self.rotation = 0 #Stores rotation of the building
        self.mined = False #Building can mine only once in logic cycle

        self.cargo = defaultdict(lambda: 0)
        self.cargo_capacity = cargo_capacity
        self.cargo_filled = 0 #How much cargo is already in

        self.time_to_finish_building = -1 #Stores time required to finish building unit
        self.building_output = []
        
        self.time_to_finish_processing = -1 #Stores time required to finish processing
        self.processing_output = []

        self.time_to_finish = time_to_finish #Time required to finish this building
        self.tool_to_build = tool_to_build #Tool required to finish this building
        if self.time_to_finish > 0:
            self.disabled = True

        #Graphic effects
        self.attack_laser = [] #Stores target for attack laser that is drawn when entity attacks

        self.sendDataToManager()

    def sendDataToManager(self):
        self.parent.all_buildings_positions[self] = (self.x,self.y)

    def selfdraw(self):
        #Draws effects
        if self.attack_laser:
            self.parent.drawing_engine.draw_team_laser(self.x,self.y,self.attack_laser[0][0],self.attack_laser[0][1],self.parent.parent.team_number)
            self.attack_laser.pop()
        
        self.parent.drawing_engine.draw_team_object(self.name,self.x,self.y,self.parent.parent.team_number,self.rotation)
        
    def update(self): #Building updates itself, called every logic cycle
        self.reload += 1

        self.mined = False
        
        if self.time_to_finish_building > 0:
            self.time_to_finish_building -= 1
        if self.time_to_finish_building == 0:
            self.finish_building()
            
        if self.time_to_finish_processing > 0:
            self.time_to_finish_processing -= 1
        if self.time_to_finish_processing == 0:
            self.finish_processing()

        if self.time_to_finish < 1 and not self.time_to_finish < -5000:
            self.disabled = False
            self.time_to_finish = -5000
            self.tool_to_build = ""

    def fire(self,position): #Uses weapon of building to fire - position is in format (x,y)
        if self.disabled: #Building is disabled
            return
        
        if self.reload >= self.weapon[2]:
            self.reload = 0
            
            #Checks if the hit position is in range using Manhattan distance
            if self.weapon[1] < (abs(position[0]-self.x)+abs(position[1]-self.y)):
                return #The building fired too far and wasted the shot that way

            #Graphic effect
            self.attack_laser = [position] #The longer the list is the longer it is drawn (for more frames)

            #Rotates building when shooting - calculates rotation
            if abs(self.y-position[1]) > abs(self.x-position[0]):
                if self.y-position[1] < 0:
                    self.rotation = 180
                else:
                    self.rotation = 0
                    
            elif abs(self.y-position[1]) < abs(self.x-position[0]) or abs(self.y-position[1]) == abs(self.x-position[0]):
                if self.x-position[0] < 0:
                    self.rotation = 270
                else:
                    self.rotation = 90
            
            #Searches for entity located at that position and then damages it
            for entity, entity_position in self.parent.parent.blocked_positions.items():
                if entity_position == position:
                    entity.take_damage(self.weapon[0])

    def take_damage(self,damage):
        self.hp -= damage

    def build(self,position,target_object,recipe_number): #@todo add support for building buildings
        if self.disabled: #Building is disabled
            return
        
        requirements = self.parent.parent.recipe_manager.recipes[target_object][recipe_number] #Obtains requirements for the target object
        if not self.tool[1] in requirements[0]: #Checks if the unit has required building tool
            return

        for i in range(len(requirements[1])): #Iterates through resources required for construction
            if (requirements[2][i] and
                not self.parent.parent.resourceManager.resources[requirements[1][i]] >= int(requirements[2][i])): #Checks if the team has enough resources
                return #If not, cancels the building order

        #Remove used amount of resources from resource manager
        for i in range(len(requirements[1])): #Iterates through resources required for construction
            if requirements[2][i]:
                self.parent.parent.resourceManager.resources[requirements[1][i]] -= int(requirements[2][i])

        self.building_output = (position[0],position[1],target_object) #Requirements have been fulfiled and the unit is being built
        self.time_to_finish_building = requirements[3]

    def finish_building(self):
        if (abs(self.x-self.building_output[0])+abs(self.y-self.building_output[1])) != 1: #Buildings can build only on tiles next to them
            return

        if ((self.building_output[0] >= len(self.parent.mapa) or self.building_output[0] <= -1) or #Checks if the target won't be built outside of the map
            (self.building_output[1] >= len(self.parent.mapa[0]) or self.building_output[1] <= -1)):
            return

        if not self.parent.mapa[self.building_output[0]][self.building_output[1]]: #Checks if the tile terrain is useable for construction (it must be passable)
            return

        if ((self.building_output[0],self.building_output[1]) in self.parent.parent.unitManager.all_units_grid_positions.values() or #Checks if there isn't entity blocking tile
            (self.building_output[0],self.building_output[1]) in self.parent.parent.blocked_positions):
            return

        self.parent.parent.create_unit(self.building_output[0],self.building_output[1],self.building_output[2])
        self.time_to_finish_building = -1
        

    def mine(self,position): #Mines from given position if possible
        if self.disabled: #Building is disabled
            return
        
        if self.mined: #Building has already mined during this logic cycle
            return

        #Graphic effect
        self.attack_laser = [position,"duration","duration"] #The longer the list is the longer it is drawn (for more frames)

        self.mined = True
        
        try:
            ore_name, ore_amount = self.parent.parent.parent.ore_manager.ores[(position[0],position[1])]
        except KeyError: #The ore has been mined out
            return

        if not self.tool[0] in self.parent.parent.parent.ore_manager.ore_types[ore_name]: #Building doesn't have useable tool for mining the ore
            return

        mined_amount = int(math.ceil(ore_amount/10.0)) #Mines 10% of ore rounding up, that means the richer the ore is, the faster it is mined
        
        if self.cargo_capacity - self.cargo_filled < mined_amount: #If there is less cargo capacity free than should be mined out, the obtained ore amount is reduced
            mined_amount = self.cargo_capacity - self.cargo_filled

        #Loads ore into cargo
        self.cargo[ore_name] = self.cargo[ore_name] + mined_amount
        self.cargo_filled += mined_amount

        #Removes mined out ore amount from deposit
        self.parent.parent.parent.ore_manager.ores[(position[0],position[1])] = (ore_name,ore_amount-mined_amount)

    def unload_cargo(self,position,cargo_name,cargo_amount): #Unloads cargo to unit or building of the same team
        if self.disabled: #Building is disabled
            return

        #Checks if the unload point is close enough
        if not (abs(position[0]-self.x)+abs(position[1]-self.y)) <= 1:
            return
        
        if (position[0],position[1]) in self.parent.all_units_grid_positions.values(): #Searches for units
            for unit, unit_position in self.parent.all_units_grid_positions.items():
                if (position[0],position[1]) == unit_position:
                    target = unit                
        elif (position[0],position[1]) in self.parent.parent.buildingManager.all_buildings_positions.values(): #Searches for buildings
            for building, building_position in self.parent.parent.buildingManager.all_buildings_positions.items():
                if (position[0],position[1]) == building_position:
                    target = building
        else: #Nothing was found
            return

        #Decreases offloaded cargo amount if there isn't enough space in target or there isn't enough of cargo
        if self.cargo[cargo_name] < cargo_amount:
            cargo_amount = self.cargo[cargo_name]

        cargo_amount = min([cargo_amount,target.cargo_capacity - target.cargo_filled])

        target.cargo[cargo_name] = target.cargo[cargo_name] + cargo_amount #Puts cargo into target
        target.cargo_filled += cargo_amount
        
        self.cargo[cargo_name] = self.cargo[cargo_name] - cargo_amount #Removes cargo from here
        self.cargo_filled -= cargo_amount

    def process_resource(self,target_resource,recipe_number): #Creates resource out of other resources
        if self.disabled: #Building is disabled
            return
        
        requirements = self.parent.parent.recipe_manager.resource_recipes[target_resource][recipe_number] #Obtains requirements for the target recipe
        if not self.tool[3] in requirements[0]: #Checks if the unit has required processing tool
            return

        for i in range(len(requirements[1])): #Iterates through resources required for processing
            if (requirements[2][i] and
                not self.parent.parent.resourceManager.resources[requirements[1][i]] >= int(requirements[2][i]) and
                not self.cargo[requirements[1][i]] >= int(requirements[2][i])): #Checks if the team or cargo has enough resources
                return #If not, cancels the building order

        #Remove used amount of resources from resource manager or from cargo
        for i in range(len(requirements[1])): #Iterates through resources required for processing
            if requirements[2][i]:
                if self.cargo[requirements[1][i]] >= int(requirements[2][i]):
                    self.cargo[requirements[1][i]] -= int(requirements[2][i])
                    self.cargo_filled -= int(requirements[2][i])
                elif self.parent.parent.resourceManager.resources[requirements[1][i]] >= int(requirements[2][i]):
                    self.parent.parent.resourceManager.resources[requirements[1][i]] -= int(requirements[2][i])

        self.processing_output = (target_resource,requirements[3])
        self.time_to_finish_processing = requirements[4]

    def finish_processing(self):
        self.parent.parent.resourceManager.resources[self.processing_output[0]] += self.processing_output[1]
        self.time_to_finish_processing = -1


class BuildingManager():
    def __init__(self, parent, mapa, drawing_engine):
        self.parent = parent
        self.mapa = mapa

        self.all_buildings = []
        self.all_buildings_positions = {}

        self.drawing_engine = drawing_engine

        #Identifier system used for commanding the buildings
        self.id_system = {}
        self.current_identifier_number = 0

    def update(self):
        #First, destroyed buildings get removed
        for i in range(len(self.all_buildings)-1,-1,-1): #Iterates from back to front what enables removal of buildings during it
            if not self.all_buildings[i].hp > 0:
                #Removes all references to the building
                del self.all_buildings_positions[self.all_buildings[i]]
                del self.id_system[self.all_buildings[i].id]
                self.all_buildings.pop(i)
                
        for i in self.all_buildings:
            i.update()

        return self.all_buildings_positions

    def draw_all_buildings(self):
        for i in self.all_buildings:
            i.selfdraw()

    def detect_building_using_position(self,x,y):
        for building in self.all_buildings:
            if building.x == x and building.y == y:
                return building
        return False

    def create_building(self, x, y, building_type, time_to_finish = 0, tool_to_build = ""):
        hp, weapon_type, tool_type, cargo_capacity = self.parent.UBWTTypes.building_types[building_type]
        weapon = self.parent.UBWTTypes.weapon_types[weapon_type]
        tool = self.parent.UBWTTypes.tool_types[tool_type]
        new_building = BasicBuilding(self, building_type, x, y, [1,1], hp, weapon, tool, cargo_capacity, time_to_finish, tool_to_build)
        self.all_buildings.append(new_building)

        #Adds unit into ID system
        self.id_system[self.current_identifier_number] = new_building
        new_building.id = self.current_identifier_number
        self.current_identifier_number += 1







    
