#Drawing module - every unit and building will be drawn using this
import pygame
#Color definition
black    = (   0,   0,   0)
white    = ( 255, 255, 255)
green    = (   0, 255,   0)
red      = ( 255,   0,   0)
blue     = (   0,   0, 255)
violet   = ( 128,   0, 127)
yellow   = ( 255, 255,   0)
grass    = (   0, 214,   0)
brown    = ( 166, 124,  35)
water    = (  20, 154, 181)

from simple_modules import moduleFilePathFormer #Necessary to access images in different folders

class DrawingEngine():
    def __init__(self, kamera, screen, tile_size, team_count, terrain_map):
        self.kamera = kamera #Used to move screen
        self.screen = screen
        self.tile_size = tile_size
        self.team_count = team_count

        self.filePathFormer = moduleFilePathFormer.FilePathFormer("Game_folder")

        self.images = {} #Stores object names (units, buildings, ores...) paired with dictionaries storing rotation
        #in degrees paired with images

        self.team_color_covers = [] #Stores semi-transparent surfaces that are blitted on the images of units and buildings
        self.create_team_color_covers()

        self.load_images()

        #Terrain
        self.terrain_map = terrain_map
        self.terrain_surface = ""
        self.create_terrain_surface()

    def create_terrain_surface(self): #Creates terrain surface to be blitted every frame
        self.terrain_surface = pygame.Surface((len(self.terrain_map)*self.tile_size,len(self.terrain_map[0])*self.tile_size))

        #Standard drawing of terrain tiles using stored images - terain tiles are called "terrain_tile_0", "terrain_tile_1" and so on
        try: 
            for i in range(len(self.terrain_map)):
                for j in range(len(self.terrain_map[i])):
                    #Blits the terrain tile with rotation 0
                    self.terrain_surface.blit(self.images["terrain_tile_"+str(self.terrain_map[i][j])][0],(i*self.tile_size,j*self.tile_size))
                    
        #Fallback drawing terrain tiles
        except KeyError:
            for i in range(len(self.terrain_map)):
                for j in range(len(self.terrain_map[i])):
                    if self.terrain_map[i][j]:
                        pygame.draw.rect(self.terrain_surface,red,[i*self.tile_size,j*self.tile_size,self.tile_size,self.tile_size],2)
                    else:
                        pygame.draw.rect(self.terrain_surface,red,[i*self.tile_size,j*self.tile_size,self.tile_size,self.tile_size])
        

    def create_team_color_covers(self):
        self.team_colors = []
        mini_color = [1,1,1] #Used so each team will always get its color and it will be constant
        for i in range(self.team_count):
            color = [mini_color[0]*50,mini_color[1]*50,mini_color[2]*50] #Multiplies mini color by 50
            self.team_colors.append(color)
            
            color_cover = pygame.Surface((self.tile_size,self.tile_size))
            color_cover.fill([color[0],color[1],color[2]])
            color_cover.set_alpha(100)
            self.team_color_covers.append(color_cover)

            #Change mini color
            mini_color[2] += 1
            if mini_color[2] > 5:
                mini_color[2] = 0
                mini_color[1] += 1
            if mini_color[1] > 5:
                mini_color[1] = 0
                mini_color[0] += 1
        
    def load_images(self): #Warning, this might take some time if there are too many images        
        file = open(self.filePathFormer.create_path_to_file(["Images","image_link.txt"]),"r") #Opens linking file
        for line in file:
            if not line[0] == "#":
                if "\n" in line: #Removes line ending
                    line = line[:-1]
                line = line.split(":")

                try: #Tries loading the image
                    image = pygame.image.load(self.filePathFormer.create_path_to_file(["Images",line[1]])).convert_alpha()
                    image = pygame.transform.scale(image,(self.tile_size,self.tile_size))
                    self.images[line[0]] = {0:image}
                except pygame.error: #Image wasn't found, replacement is generated
                    replacement = pygame.Surface((self.tile_size,self.tile_size))
                    pygame.draw.rect(replacement,violet,[0,0,self.tile_size/2,self.tile_size/2])
                    pygame.draw.rect(replacement,black,[self.tile_size/2,0,self.tile_size/2,self.tile_size/2])
                    pygame.draw.rect(replacement,black,[0,self.tile_size/2,self.tile_size/2,self.tile_size/2])
                    pygame.draw.rect(replacement,violet,[self.tile_size/2,self.tile_size/2,self.tile_size/2,self.tile_size/2])
                    self.images[line[0]] = {0:replacement}

        file.close()

    def draw_terrain(self):
        self.screen.blit(self.terrain_surface,(-self.kamera.x,-self.kamera.y))

    def draw_object(self,object_name,x,y,rotation=0): #Used for drawing neutral objects like ores
        #Rotation is in degrees counter clockwise and must be divisible by 90
        x = x*self.tile_size
        y = y*self.tile_size
        
        try: #If the rotated image isn't yet saved, rotate the unrotated image
            image_to_be_blitted = self.images[object_name][rotation]
        except KeyError:
            self.images[object_name][rotation] = pygame.transform.rotate(self.images[object_name][0],rotation)
            image_to_be_blitted = self.images[object_name][rotation]

        draw_x = x - self.kamera.x
        draw_y = y - self.kamera.y

        self.screen.blit(image_to_be_blitted,(draw_x,draw_y))

    def draw_team_cover(self,x,y,team): #No longer used
        x = x*self.tile_size
        y = y*self.tile_size
        draw_x = x - self.kamera.x
        draw_y = y - self.kamera.y
        
        self.screen.blit(self.team_color_covers[team],(draw_x,draw_y))

    def draw_team_object(self,object_name,x,y,team,rotation=0): #Used for drawing units and buildings
        #Rotation is in degrees counter clockwise and must be divisible by 90
        x = x*self.tile_size
        y = y*self.tile_size
        draw_x = x - self.kamera.x
        draw_y = y - self.kamera.y
        
        try: #If the rotated image isn't yet saved, rotate the unrotated image
            image_to_be_blitted = self.images[object_name][rotation]
        except KeyError:
            self.images[object_name][rotation] = pygame.transform.rotate(self.images[object_name][0],rotation)
            image_to_be_blitted = self.images[object_name][rotation]

        temporary_image = image_to_be_blitted.copy() #Copies the object image and then colors it using team color cover
        temporary_image.blit(self.team_color_covers[team],(0,0),special_flags = pygame.BLEND_MULT)
        
        self.screen.blit(temporary_image,(draw_x,draw_y))

    def draw_team_laser(self,origin_x,origin_y,target_x,target_y,team): #Used for drawing team colored interaction laser
        origin_x = origin_x*self.tile_size+self.tile_size/2
        origin_y = origin_y*self.tile_size+self.tile_size/2
        origin_draw_x = origin_x - self.kamera.x
        origin_draw_y = origin_y - self.kamera.y

        target_x = target_x*self.tile_size+self.tile_size/2
        target_y = target_y*self.tile_size+self.tile_size/2
        target_draw_x = target_x - self.kamera.x
        target_draw_y = target_y - self.kamera.y

        pygame.draw.line(self.screen,self.team_colors[team],(origin_draw_x,origin_draw_y),(target_draw_x,target_draw_y),3)

    def draw_side_bar(self): #Used to draw side bar that is part of GUI
        screen_size = self.screen.get_size()
        pygame.draw.rect(self.screen,black,[screen_size[0]/10*9,0,screen_size[0],screen_size[1]])
        pygame.draw.line(self.screen,blue,[screen_size[0]/10*9-5/2,0],[screen_size[0]/10*9-5/2,screen_size[1]],5)
    


        
        
